@@ -12,16 +12,14 @@ import random
 import typing
 from dataclasses import dataclass
 
-from terminaltexteffects.engine.base_character import EffectCharacter, EventHandler
+from terminaltexteffects import Color, ColorPair, Coord, EffectCharacter, EventHandler, Gradient, Terminal, geometry
 from terminaltexteffects.engine.base_effect import BaseEffect, BaseEffectIterator
-from terminaltexteffects.engine.terminal import Terminal
-from terminaltexteffects.utils import argvalidators, geometry
+from terminaltexteffects.utils import argvalidators
 from terminaltexteffects.utils.argsdataclass import ArgField, ArgsDataClass, argclass
-from terminaltexteffects.utils.geometry import Coord
-from terminaltexteffects.utils.graphics import Color, Gradient
 
 
 def get_effect_and_args() -> tuple[type[typing.Any], type[ArgsDataClass]]:
+    """Get the effect class and its configuration class."""
     return SynthGrid, SynthGridConfig
 
 
@@ -29,7 +27,11 @@ def get_effect_and_args() -> tuple[type[typing.Any], type[ArgsDataClass]]:
     name="synthgrid",
     help="Create a grid which fills with characters dissolving into the final text.",
     description="synthgrid | Create a grid which fills with characters dissolving into the final text.",
-    epilog="""Example: terminaltexteffects synthgrid --grid-gradient-stops CC00CC ffffff --grid-gradient-steps 12 --text-gradient-stops 8A008A 00D1FF FFFFFF --text-gradient-steps 12 --grid-row-symbol ─ --grid-column-symbol "│" --text-generation-symbols ░ ▒ ▓ --max-active-blocks 0.1""",
+    epilog=(
+        "Example: terminaltexteffects synthgrid --grid-gradient-stops CC00CC ffffff --grid-gradient-steps 12 "
+        "--text-gradient-stops 8A008A 00D1FF FFFFFF --text-gradient-steps 12 --grid-row-symbol ─ "
+        "--grid-column-symbol '|' --text-generation-symbols ░ ▒ ▓ --max-active-blocks 0.1"
+    ),
 )
 @dataclass
 class SynthGridConfig(ArgsDataClass):
@@ -37,15 +39,20 @@ class SynthGridConfig(ArgsDataClass):
 
     Attributes:
         grid_gradient_stops (tuple[Color, ...]): Tuple of colors for the grid gradient.
-        grid_gradient_steps (tuple[int, ...]): Tuple of the number of gradient steps to use. More steps will create a smoother and longer gradient animation. Valid values are n > 0.
+        grid_gradient_steps (tuple[int, ...] | int ): Int or Tuple of the number of gradient steps to use.
+            More steps will create a smoother and longer gradient animation. Valid values are n > 0.
         grid_gradient_direction (Gradient.Direction): Direction of the gradient for the grid color.
         text_gradient_stops (tuple[Color, ...]): Tuple of colors for the text gradient.
-        text_gradient_steps (tuple[int, ...]): Tuple of the number of gradient steps to use. More steps will create a smoother and longer gradient animation. Valid values are n > 0.
+        text_gradient_steps (tuple[int, ...] | int ): Int or Tuple of the number of gradient steps to use. More
+            steps will create a smoother and longer gradient animation. Valid values are n > 0.
         text_gradient_direction (Gradient.Direction): Direction of the gradient for the text color.
         grid_row_symbol (str): Symbol to use for grid row lines.
         grid_column_symbol (str): Symbol to use for grid column lines.
         text_generation_symbols (tuple[str, ...] | str): Tuple of characters for the text generation animation.
-        max_active_blocks (float): Maximum percentage of blocks to have active at any given time. For example, if set to 0.1, 10 percent of the blocks will be active at any given time. Valid values are 0 < n <= 1."""
+        max_active_blocks (float): Maximum percentage of blocks to have active at any given time. For example,
+            if set to 0.1, 10 percent of the blocks will be active at any given time. Valid values are 0 < n <= 1.
+
+    """
 
     grid_gradient_stops: tuple[Color, ...] = ArgField(
         cmd_name=["--grid-gradient-stops"],
@@ -63,9 +70,13 @@ class SynthGridConfig(ArgsDataClass):
         nargs="+",
         default=12,
         metavar=argvalidators.PositiveInt.METAVAR,
-        help="Space separated, unquoted, list of the number of gradient steps to use. More steps will create a smoother and longer gradient animation.",
+        help="Space separated, unquoted, list of the number of gradient steps to use. More steps will create a "
+        "smoother and longer gradient animation.",
     )  # type: ignore[assignment]
-    "tuple[int, ...] | int : Int or Tuple of ints for the number of gradient steps to use. More steps will create a smoother and longer gradient animation."
+    (
+        "tuple[int, ...] | int : Int or Tuple of ints for the number of gradient steps to use. More steps will "
+        "create a smoother and longer gradient animation."
+    )
 
     grid_gradient_direction: Gradient.Direction = ArgField(
         cmd_name="--grid-gradient-direction",
@@ -92,9 +103,13 @@ class SynthGridConfig(ArgsDataClass):
         nargs="+",
         default=12,
         metavar=argvalidators.PositiveInt.METAVAR,
-        help="Space separated, unquoted, list of the number of gradient steps to use. More steps will create a smoother and longer gradient animation.",
+        help="Space separated, unquoted, list of the number of gradient steps to use. More steps will create a "
+        "smoother and longer gradient animation.",
     )  # type: ignore[assignment]
-    "tuple[int, ...] | int : Int or Tuple of ints for the number of gradient steps to use. More steps will create a smoother and longer gradient animation."
+    (
+        "tuple[int, ...] | int : Int or Tuple of ints for the number of gradient steps to use. More steps will "
+        "create a smoother and longer gradient animation."
+    )
 
     text_gradient_direction: Gradient.Direction = ArgField(
         cmd_name="--text-gradient-direction",
@@ -135,19 +150,23 @@ class SynthGridConfig(ArgsDataClass):
 
     max_active_blocks: float = ArgField(
         cmd_name="--max-active-blocks",
-        type_parser=argvalidators.PositiveFloat.type_parser,
+        type_parser=argvalidators.PositiveRatio.type_parser,
         default=0.1,
-        metavar=argvalidators.PositiveFloat.METAVAR,
-        help="Maximum percentage of blocks to have active at any given time. For example, if set to 0.1, 10 percent of the blocks will be active at any given time.",
+        metavar=argvalidators.PositiveRatio.METAVAR,
+        help="Maximum percentage of blocks to have active at any given time. For example, if set to 0.1, 10 percent "
+        "of the blocks will be active at any given time.",
     )  # type: ignore[assignment]
     "float : Maximum percentage of blocks to have active at any given time."
 
     @classmethod
-    def get_effect_class(cls):
+    def get_effect_class(cls) -> type[SynthGrid]:
+        """Get the effect class associated with this configuration."""
         return SynthGrid
 
 
 class GridLine:
+    """A line in the grid."""
+
     def __init__(
         self,
         terminal: Terminal,
@@ -155,7 +174,17 @@ class GridLine:
         origin: Coord,
         direction: str,
         grid_gradient_mapping: dict[geometry.Coord, Color],
-    ):
+    ) -> None:
+        """Initialize the grid line.
+
+        Args:
+            terminal (Terminal): Terminal from the effect.
+            args (SynthGridConfig): Configuration for the effect.
+            origin (Coord): Origin coordinate.
+            direction (str): Direction of the line.
+            grid_gradient_mapping (dict[geometry.Coord, Color]): Mapping of coordinates to colors.
+
+        """
         self.terminal = terminal
         self.args = args
         self.origin = origin
@@ -170,7 +199,9 @@ class GridLine:
                 effect_char = self.terminal.add_character(self.grid_symbol, Coord(0, 0))
                 grid_scn = effect_char.animation.new_scene()
                 grid_scn.add_frame(
-                    self.grid_symbol, 1, color=grid_gradient_mapping[geometry.Coord(column_index, origin.row)]
+                    self.grid_symbol,
+                    1,
+                    colors=ColorPair(fg=grid_gradient_mapping[geometry.Coord(column_index, origin.row)]),
                 )
                 effect_char.animation.activate_scene(grid_scn)
                 effect_char.layer = 2
@@ -181,48 +212,66 @@ class GridLine:
                 effect_char = self.terminal.add_character(self.grid_symbol, Coord(0, 0))
                 grid_scn = effect_char.animation.new_scene()
                 grid_scn.add_frame(
-                    self.grid_symbol, 1, color=grid_gradient_mapping[geometry.Coord(origin.column, row_index)]
+                    self.grid_symbol,
+                    1,
+                    colors=ColorPair(fg=grid_gradient_mapping[geometry.Coord(origin.column, row_index)]),
                 )
                 effect_char.animation.activate_scene(grid_scn)
                 effect_char.layer = 2
                 effect_char.motion.set_coordinate(Coord(origin.column, row_index))
                 self.characters.append(effect_char)
-        self.collapsed_characters = [effect_char for effect_char in self.characters]
+        self.collapsed_characters = list(self.characters)
         self.extended_characters: list[EffectCharacter] = []
 
     def extend(self) -> None:
-        if self.direction == "horizontal":
-            count = 3
-        else:
-            count = 1
+        """Extend the line."""
+        count = 3 if self.direction == "horizontal" else 1
         for _ in range(count):
             if self.collapsed_characters:
                 next_char = self.collapsed_characters.pop(0)
-                self.terminal.set_character_visibility(next_char, True)
+                self.terminal.set_character_visibility(next_char, is_visible=True)
                 self.extended_characters.append(next_char)
 
     def collapse(self) -> None:
-        if self.direction == "horizontal":
-            count = 3
-        else:
-            count = 1
+        """Collapse the line."""
+        count = 3 if self.direction == "horizontal" else 1
         if not self.collapsed_characters:
             self.extended_characters = self.extended_characters[::-1]
         for _ in range(count):
             if self.extended_characters:
                 next_char = self.extended_characters.pop(0)
-                self.terminal.set_character_visibility(next_char, False)
+                self.terminal.set_character_visibility(next_char, is_visible=False)
                 self.collapsed_characters.append(next_char)
 
     def is_extended(self) -> bool:
+        """Check if the line is extended.
+
+        Returns:
+            bool: True if the line is extended, False otherwise.
+
+        """
         return not self.collapsed_characters
 
     def is_collapsed(self) -> bool:
+        """Check if the line is collapsed.
+
+        Returns:
+            bool: True if the line is collapsed, False otherwise.
+
+        """
         return not self.extended_characters
 
 
 class SynthGridIterator(BaseEffectIterator[SynthGridConfig]):
-    def __init__(self, effect: "SynthGrid") -> None:
+    """Iterator for the SynthGrid effect."""
+
+    def __init__(self, effect: SynthGrid) -> None:
+        """Initialize the effect iterator.
+
+        Args:
+            effect (SynthGrid): The effect to use for the iterator.
+
+        """
         super().__init__(effect)
         self.pending_groups: list[tuple[int, list[EffectCharacter]]] = []
         self.grid_lines: list[GridLine] = []
@@ -237,26 +286,33 @@ class SynthGridIterator(BaseEffectIterator[SynthGridConfig]):
 
         Returns:
             int: The gap that is closest to 20% of the dimension length.
+
         """
-        potential_gaps: list[int] = []
         dimension = dimension - 2
         if dimension <= 0:
             return 0
-        for i in range(dimension, 4, -1):
-            if dimension % i <= 1:
-                potential_gaps.append(i)
+        potential_gaps: list[int] = [i for i in range(dimension, 4, -1) if dimension % i <= 1]
         if not potential_gaps:
             return 4
         return min(potential_gaps, key=lambda x: abs(x - dimension // 5))
 
-    def build(self) -> None:
+    def build(self) -> None:  # noqa: PLR0915
+        """Build the initial state of the effect."""
         grid_gradient = Gradient(*self.config.grid_gradient_stops, steps=self.config.grid_gradient_steps)
         grid_gradient_mapping = grid_gradient.build_coordinate_color_mapping(
-            self.terminal.canvas.top, self.terminal.canvas.right, self.config.grid_gradient_direction
+            1,
+            self.terminal.canvas.top,
+            1,
+            self.terminal.canvas.right,
+            self.config.grid_gradient_direction,
         )
         text_gradient = Gradient(*self.config.text_gradient_stops, steps=self.config.text_gradient_steps)
         text_gradient_mapping = text_gradient.build_coordinate_color_mapping(
-            self.terminal.canvas.top, self.terminal.canvas.right, self.config.text_gradient_direction
+            self.terminal.canvas.text_bottom,
+            self.terminal.canvas.text_top,
+            self.terminal.canvas.text_left,
+            self.terminal.canvas.text_right,
+            self.config.text_gradient_direction,
         )
 
         self.grid_lines.append(
@@ -266,7 +322,7 @@ class SynthGridIterator(BaseEffectIterator[SynthGridConfig]):
                 Coord(self.terminal.canvas.left, self.terminal.canvas.bottom),
                 "horizontal",
                 grid_gradient_mapping,
-            )
+            ),
         )
         self.grid_lines.append(
             GridLine(
@@ -275,7 +331,7 @@ class SynthGridIterator(BaseEffectIterator[SynthGridConfig]):
                 Coord(self.terminal.canvas.left, self.terminal.canvas.top),
                 "horizontal",
                 grid_gradient_mapping,
-            )
+            ),
         )
         self.grid_lines.append(
             GridLine(
@@ -284,7 +340,7 @@ class SynthGridIterator(BaseEffectIterator[SynthGridConfig]):
                 Coord(self.terminal.canvas.left, self.terminal.canvas.bottom),
                 "vertical",
                 grid_gradient_mapping,
-            )
+            ),
         )
         self.grid_lines.append(
             GridLine(
@@ -293,7 +349,7 @@ class SynthGridIterator(BaseEffectIterator[SynthGridConfig]):
                 Coord(self.terminal.canvas.right, self.terminal.canvas.bottom),
                 "vertical",
                 grid_gradient_mapping,
-            )
+            ),
         )
         column_indexes: list[int] = []
         row_indexes: list[int] = []
@@ -315,10 +371,12 @@ class SynthGridIterator(BaseEffectIterator[SynthGridConfig]):
                     Coord(self.terminal.canvas.left, row_index),
                     "horizontal",
                     grid_gradient_mapping,
-                )
+                ),
             )
         for column_index in range(
-            self.terminal.canvas.left + column_gap, self.terminal.canvas.right, max(column_gap, 1)
+            self.terminal.canvas.left + column_gap,
+            self.terminal.canvas.right,
+            max(column_gap, 1),
         ):
             if self.terminal.canvas.right - column_index < 2:
                 continue
@@ -330,7 +388,7 @@ class SynthGridIterator(BaseEffectIterator[SynthGridConfig]):
                     Coord(column_index, self.terminal.canvas.bottom),
                     "vertical",
                     grid_gradient_mapping,
-                )
+                ),
             )
         row_indexes.append(self.terminal.canvas.top + 1)
         column_indexes.append(self.terminal.canvas.right + 1)
@@ -340,14 +398,15 @@ class SynthGridIterator(BaseEffectIterator[SynthGridConfig]):
             for column_index in column_indexes:
                 coords_in_block: list[Coord] = []
                 if row_index == self.terminal.canvas.top:  # make sure the top row is included
-                    row_index += 1
+                    row_index += 1  # noqa: PLW2901
                 for row in range(prev_row_index, row_index):
                     for column in range(prev_column_index, column_index):
-                        coords_in_block.append(Coord(column, row))
-                characters_in_block: list[EffectCharacter] = []
-                for coord in coords_in_block:
-                    if coord in self.terminal.character_by_input_coord:
-                        characters_in_block.append(self.terminal.character_by_input_coord[coord])
+                        coords_in_block.append(Coord(column, row))  # noqa: PERF401
+                characters_in_block: list[EffectCharacter] = [
+                    self.terminal.character_by_input_coord[coord]
+                    for coord in coords_in_block
+                    if coord in self.terminal.character_by_input_coord
+                ]
                 if characters_in_block:
                     self.pending_groups.append((len(self.pending_groups), characters_in_block))
                 prev_column_index = column_index
@@ -360,9 +419,16 @@ class SynthGridIterator(BaseEffectIterator[SynthGridConfig]):
                     dissolve_scn.add_frame(
                         random.choice(self.config.text_generation_symbols),
                         3,
-                        color=random.choice(text_gradient.spectrum),
+                        colors=ColorPair(fg=random.choice(text_gradient.spectrum)),
+                    )
+                if character.input_symbol == " ":
+                    dissolve_scn.add_frame(character.input_symbol, 1)
+                else:
+                    dissolve_scn.add_frame(
+                        character.input_symbol,
+                        1,
+                        colors=ColorPair(fg=text_gradient_mapping[character.input_coord]),
                     )
-                dissolve_scn.add_frame(character.input_symbol, 1, color=text_gradient_mapping[character.input_coord])
                 character.animation.activate_scene(dissolve_scn)
                 character.event_handler.register_event(
                     EventHandler.Event.SCENE_COMPLETE,
@@ -375,17 +441,19 @@ class SynthGridIterator(BaseEffectIterator[SynthGridConfig]):
         self._total_group_count = len(self.pending_groups)
         if not self._total_group_count:
             for character in self.terminal.get_characters():
-                self.terminal.set_character_visibility(character, True)
-                self.active_characters.append(character)
+                self.terminal.set_character_visibility(character, is_visible=True)
+                self.active_characters.add(character)
         self._active_groups: int = 0
 
-    def update_group_tracker(self, character: EffectCharacter, *args) -> None:
+    def update_group_tracker(self, _: EffectCharacter, *args) -> None:  # noqa: ANN002
+        """Update the group tracker."""
         self.group_tracker[args[0]] -= 1
 
     def __next__(self) -> str:
+        """Return the next frame in the animation."""
         if self.pending_groups or self.active_characters or self._phase != "complete":
             if self._phase == "grid_expand":
-                if not all([grid_line.is_extended() for grid_line in self.grid_lines]):
+                if not all(grid_line.is_extended() for grid_line in self.grid_lines):
                     for grid_line in self.grid_lines:
                         if not grid_line.is_extended():
                             grid_line.extend()
@@ -398,13 +466,13 @@ class SynthGridIterator(BaseEffectIterator[SynthGridConfig]):
                 ):
                     group_number, next_group = self.pending_groups.pop(0)
                     for char in next_group:
-                        self.terminal.set_character_visibility(char, True)
-                        self.active_characters.append(char)
+                        self.terminal.set_character_visibility(char, is_visible=True)
+                        self.active_characters.add(char)
                         self.group_tracker[group_number] += 1
                 if not self.pending_groups and not self.active_characters and not self._active_groups:
                     self._phase = "collapse"
             elif self._phase == "collapse":
-                if not all([grid_line.is_collapsed() for grid_line in self.grid_lines]):
+                if not all(grid_line.is_collapsed() for grid_line in self.grid_lines):
                     for grid_line in self.grid_lines:
                         if not grid_line.is_collapsed():
                             grid_line.collapse()
@@ -412,12 +480,11 @@ class SynthGridIterator(BaseEffectIterator[SynthGridConfig]):
                     self._phase = "complete"
             self.update()
             self._active_groups = 0
-            for _, active_count in self.group_tracker.items():
+            for active_count in self.group_tracker.values():
                 if active_count:
                     self._active_groups += 1
             return self.frame
-        else:
-            raise StopIteration
+        raise StopIteration
 
 
 class SynthGrid(BaseEffect[SynthGridConfig]):
@@ -426,14 +493,13 @@ class SynthGrid(BaseEffect[SynthGridConfig]):
     Attributes:
         effect_config (SynthGridConfig): Configuration for the effect.
         terminal_config (TerminalConfig): Configuration for the terminal.
+
     """
 
-    _config_cls = SynthGridConfig
-    _iterator_cls = SynthGridIterator
+    @property
+    def _config_cls(self) -> type[SynthGridConfig]:
+        return SynthGridConfig
 
-    def __init__(self, input_data: str) -> None:
-        """Initialize the effect with the provided input data.
-
-        Args:
-            input_data (str): The input data to use for the effect."""
-        super().__init__(input_data)
+    @property
+    def _iterator_cls(self) -> type[SynthGridIterator]:
+        return SynthGridIterator
